// example.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "../s3.h"


//      
class FileSoundSource : public s3::Source
{
	FILE *f;

public:
	FileSoundSource(const char *name) : Source(s3::SG_MUSIC) { f = fopen(name, "rb"); }
	~FileSoundSource() { Source::stop(); fclose(f); }

	/*virtual*/ int read(char *dest, int size) { return fread(dest, 1, size, f); }
	/*virtual*/ void rewind(bool start) { if (!start) ::rewind(f); }
};


//     
class SimpleMSource : public s3::MSource
{
	char *data;//    (     )

public:
	SimpleMSource(const char *name) : MSource(s3::SG_MUSIC)
	{
		//    
		FILE *f = fopen(name, "rb");
		fseek(f, 0, SEEK_END);
		long fsize = ftell(f);
		fseek(f, 0, SEEK_SET);
		data = new char [fsize];
		fread(data, 1, fsize, f);
		fclose(f);

		init(data, fsize);
	}
	~SimpleMSource() { Source::stop(); delete [] data; }
};


//     prefetch- (    -! prefetch    -!)
class SimplePSource : public s3::PSource
{
	HANDLE file, thread;
	char *dest;
	int size;

	static DWORD WINAPI readProc(LPVOID pData) {((SimplePSource*)pData)->readFunc(); return 0;}
	void readFunc()
	{
		DWORD read = 0;
		ReadFile(file, dest, size, &read, NULL);
		if (prefetchComplete(read)) SetFilePointer(file, 0, NULL, FILE_BEGIN);
	}
	void waitForPrefetchComplete()
	{
		if (thread) WaitForSingleObject(thread, INFINITE), CloseHandle(thread), thread = NULL;
	}

public:
	SimplePSource(const char *name) : PSource(s3::SG_MUSIC, true, true), thread(NULL)
	{
		file = CreateFile(name, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	}
	~SimplePSource()
	{
		waitForPrefetchComplete();//   prefetch- (.   PSource::~PSource)
		if (file != INVALID_HANDLE_VALUE) CloseHandle(file);
	}

	/*virtual*/ void prefetch(char *dest_, int size_)
	{
		dest = dest_;
		size = size_;
		waitForPrefetchComplete();//    ,  .. update    ,  prefetchComplete -  ,  ,     readProc   
		thread = CreateThread(NULL, 0, &readProc, this, 0, NULL);
	}

	void stop(float time = 0)
	{
		waitForPrefetchComplete();//   prefetch- (.   PSource::stop)
		PSource::stop(time);
		SetFilePointer(file, 0, NULL, FILE_BEGIN);
	}
};


int main()
{
#ifdef _DEBUG
	_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF|_CRTDBG_LEAK_CHECK_DF);
#endif

	s3::Initialize();

	{
	FileSoundSource fss("sounds\\xpbatlow.flac");
	SimplePSource   sps("sounds\\utopiawi.ogg");
	sps.play();
	for (int i=0; i<450; i++)
	{
		sps.update();
		Sleep(20);
		if (i ==  50)
		{
			SimpleMSource *sms = new SimpleMSource("sounds\\ringin.wav");//  
			sms->volume = .25f;
			sms->play();
			sms->autoDelete();//   ,   !
		}
		if (i == 200) sps.stop(1);
 		if (i == 250) fss.play();
 		if (i == 300) sps.activate(), sps.play(2);
	}
	}//        s3::Shutdown()

	s3::Shutdown();

	return 0;
}
