#include "stdafx.h"
#include "slot.h"


#pragma comment(lib, "dxguid")
#pragma comment(lib, "dsound")
#pragma comment(lib, "winmm")

namespace s3
{
void DefaultErrorHandler(ErrorLevel el, const char *err)
{
	OutputDebugString(el == EL_ERROR ? "s3 ERROR: " : "s3 warning: ");
	OutputDebugString(err);
	OutputDebugString("\n");
}
void (*ErrorHandler)(ErrorLevel, const char *) = &DefaultErrorHandler;

LPDIRECTSOUND8 pDS = NULL;
HWND hwnd = NULL;
LPDIRECTSOUND3DLISTENER8 dsListener = NULL;
LPDIRECTSOUNDBUFFER dsPrimaryBuffer = NULL;
HANDLE hQuitEvent = NULL, hUpdateThread = NULL;
CRITICAL_SECTION critSect;
InitParams initParams;//,     (  )
int sgCount;
SoundGroupSlots *soundGroups = NULL;
Decoder *gdecoder = NULL;

SoundGroupSlots::~SoundGroupSlots()
{
	delete [] slots;
}

void SoundGroupSlots::initialize(const SlotInitParams &sip)
{
	slots = new Slot [max=sip.max];
	is3d = sip.is3d;
}

DWORD WINAPI UpdateThreadProc(LPVOID /*pData*/)
{
	DWORD prevTime = timeGetTime();

	while (WaitForSingleObject(hQuitEvent, 50) != WAIT_OBJECT_0)
	{
		AutoCriticalSection acs;

		DWORD time = timeGetTime();
		float dt = (time - prevTime)*(1/1000.f);

		for (int g=0; g<sgCount; g++)
		{
			Slot *slots = soundGroups[g].slots;
			for (int i=0, n=soundGroups[g].active; i<n; i++) slots[i].update(dt);
		}

		/*if (dsListener) */dsListener->CommitDeferredSettings();

		prevTime = time;
	}

	return 0;
}

bool Initialize(const InitParams &ip, const SlotInitParams slotsIP[], const int sgCount_)
{
	initParams = ip;
	InitializeCriticalSectionAndSpinCount(&critSect, 5000);

	// Direct Sound
	if (FAILED(DirectSoundCreate8(&DSDEVID_DefaultPlayback, &pDS, NULL))) {Shutdown(); return false;}

	//C  ..
	if (ip.hwnd == NULL)
	{
		WNDCLASS wc = {0, DefWindowProc, 0, 0, GetModuleHandle(NULL), NULL, NULL, NULL, NULL, "s3hwndclass"};
		RegisterClass(&wc);
		hwnd = CreateWindow(wc.lpszClassName, "", WS_POPUP, 0, 0, 0, 0, NULL, NULL, wc.hInstance, NULL);
	}

	//..  ,   Cooperative Level
	if (FAILED(pDS->SetCooperativeLevel(ip.hwnd ? (HWND)ip.hwnd : hwnd, DSSCL_PRIORITY))) {Shutdown(); return false;}

	// listener- (    primary buffer)
	DSBUFFERDESC dsbd = {sizeof(DSBUFFERDESC), DSBCAPS_PRIMARYBUFFER|DSBCAPS_CTRL3D|DSBCAPS_CTRLVOLUME};
	if (SUCCEEDED(pDS->CreateSoundBuffer(&dsbd, &dsPrimaryBuffer, NULL)))
	{
		dsPrimaryBuffer->QueryInterface(IID_IDirectSound3DListener8, (LPVOID*)&dsListener);

		//   primary buffer- (-    8 22)
		WAVEFORMATEX wf = BuildWaveFormat(ip.channels, ip.sampleRate, ip.bitsPerSample);
		dsPrimaryBuffer->SetFormat(&wf);
	}
	else {Shutdown(); return false;}

	//   s3
	sgCount = sgCount_;
	soundGroups = new SoundGroupSlots[sgCount];
	for (int i=0; i<sgCount; i++)
		soundGroups[i].initialize(slotsIP[i]);
	gdecoder = new Decoder;

	hQuitEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	hUpdateThread = CreateThread(NULL, 0, &UpdateThreadProc, 0, 0, NULL);
	SetThreadPriority(hUpdateThread, THREAD_PRIORITY_HIGHEST);

	return true;
}

void Shutdown()
{
	if (hUpdateThread)
	{
		SetEvent(hQuitEvent);
		WaitForSingleObject(hUpdateThread, INFINITE);
		CloseHandle(hQuitEvent);
		CloseHandle(hUpdateThread);
		hQuitEvent = hUpdateThread = NULL;
	}

	// autoDeleteSources   , ..    ,       
	for (Source *s=Source::firstSourceToDelete, *next; s; s=next)
	{
		next = s->nextSourceToDelete;
		delete s;
	}
	Source::firstSourceToDelete = NULL;

	delete [] soundGroups;
	soundGroups = NULL;
	delete gdecoder;
	gdecoder = NULL;

	if (dsListener) dsListener->Release(), dsListener = NULL;
	if (dsPrimaryBuffer) dsPrimaryBuffer->Release(), dsPrimaryBuffer = NULL;
	if (pDS) pDS->Release(), pDS = NULL;
	if (hwnd) DestroyWindow(hwnd), hwnd = NULL;

	DeleteCriticalSection(&critSect);
}

void SetMasterVolume(float f)
{
	if (dsPrimaryBuffer) dsPrimaryBuffer->SetVolume(LinearToDirectSoundVolume(f));
}

void SetListenerParameters(const float pos[3], const float front[3], const float top[3], const float vel[3], float distFactor, float rolloffFactor, float dopplerFactor)
{
	if (dsListener)
	{
		DS3DLISTENER pars;
		pars.dwSize = sizeof(DS3DLISTENER);
		ConvertVectorToDirectSoundCS(pars.vPosition   , pos);
		ConvertVectorToDirectSoundCS(pars.vOrientFront, front);
		ConvertVectorToDirectSoundCS(pars.vOrientTop  , top);
		pars.flDistanceFactor = distFactor;
		pars.flRolloffFactor = rolloffFactor;
		pars.flDopplerFactor = dopplerFactor;
		if (vel) ConvertVectorToDirectSoundCS(pars.vVelocity, vel); else memset(&pars.vVelocity, 0, sizeof(vel));
		dsListener->SetAllParameters(&pars, DS3D_DEFERRED);
	}
}
}
