#pragma once
#pragma comment(lib, "s3")

//#include "decoder.h"


//    (: SG(name, maxSlots, 3d))
#ifndef S3_SOUND_GROUPS
#define S3_SOUND_GROUPS \
	SG(SG_3D, 32, true)\
	SG(SG_MUSIC, 2, false)\
	SG(SG_UI, 5, false)\
	SG(SG_VOICE, 3, false)
#endif

namespace s3
{
enum ErrorLevel {EL_ERROR=0, EL_WARNING};
extern void (*ErrorHandler)(ErrorLevel, const char *);

extern CRITICAL_SECTION critSect;
class AutoCriticalSection
{
	LPCRITICAL_SECTION cs;

public:
	AutoCriticalSection(CRITICAL_SECTION &cs_ = critSect) : cs(&cs_) {EnterCriticalSection(cs);}
	~AutoCriticalSection() {LeaveCriticalSection(cs);}
};

#define SG(name, ...) name,
enum SoundGroup {S3_SOUND_GROUPS SG_COUNT};
#undef  SG
//typedef char SoundGroupsCountStaticCheck[SG_COUNT <= SG_MAX ? 1 : 0];

class Source//     (   MSource/PSource)
{
	friend class Slot;
	friend void Shutdown();//   firstSourceToDelete
	const SoundGroup soundGroup;
	Source &operator=(const Source &);// 
	static int readFunc(char *dest, int size, void *userPtr) {return ((Source*)userPtr)->read(dest, size);}
	static Source *firstSourceToDelete;
	Source *nextSourceToDelete;
	static void autoDeleteSources();
protected:
	int slotIndex;

public:
	float volume;//   (    ,      )
	void autoDelete();//   new       (     ),       

	Source(SoundGroup soundGroup);
	virtual ~Source() {}//*if (slotIndex != -1) */stop();}//      , ..   pure virtual call of read(), ..   .   stop()      

	virtual int read(char *dest, int size) = 0;//    dest       (0    , < 0 -   )
	virtual void rewind(bool start) {if (!start) ErrorHandler(EL_ERROR, "Can't loop sound since rewind() not implemented");}

//	void canPlayNow();
	void play(bool looping = false, float time = 0);//      , ..           (!) ,        read()!
	void stop(float time = 0);

	//3D- (     ,  ,       AutoCriticalSection)
	float position[3];
	float velocity[3];
	float minDist, maxDist;
};

class MSource : public Source//     ,    , ..    
{
	char *data;
	int size, pos;
	float startPlayOnInit;
	/*virtual*/ int read(char *dest, int size);
	/*virtual*/ void rewind(bool /*start*/) {pos = 0;}

public:
	MSource(SoundGroup soundGroup) : Source(soundGroup), startPlayOnInit(-1) {init();}
	~MSource() {Source::stop();}

	void init(void *data = NULL, int size = 0);
	void play(bool looping = false, float time = 0);
	void stop(float time = 0) {startPlayOnInit = -1; Source::stop(time);}
};

class PSource : public Source//       prefetch- (  , . )
{
	char *prefetchBuffer, *buf[2];//0 -  , 1 - 
	int bufPos, actualDataSize[2], curBuf, prBuf, eofPos[2], prefetchBytes;
	float startPlayOnPrefetchComplete;
	bool waitingForPrefetchComplete, active, eof,
		looped;//      , .. prefetch  ,    play()
	/*virtual*/ int read(char *dest, int size);
	/*virtual*/ void rewind(bool start);

public:
	PSource(SoundGroup soundGroup, bool looped = false, bool activate = false, int prefetchBytes = 0);
	~PSource() {Source::stop(); delete [] prefetchBuffer;}//   ()     ,       !

	virtual void prefetch(char *dest, int size) = 0;// ,        dest;     prefetchComplete(); ,    prefetchComplete,  prefetch   
	bool prefetchComplete(int size);// size     -     (     true -      eof,   ,   prefetch    )
	void activate();//     
	void play(float time = 0);
	void stop(float time = 0) {startPlayOnPrefetchComplete = -1; Source::stop(time); active = false;}//      prefetch (   )!
	void update();//    -    
};

class RawSource : public Source//     ,     (..    RAW-  )
{
	int readStage;
	/*virtual*/ int read(char *dest, int size);
	/*virtual*/ void rewind(bool start);

public:
	//      
	short channels, bitsPerSample;
	int sampleRate;

	RawSource(SoundGroup soundGroup) : Source(soundGroup), channels(0), bitsPerSample(0), sampleRate(0) {}
	~RawSource() {Source::stop();}

	void play(bool looping = false, float time = 0) {Source::play(looping, time);}

	virtual int rawRead(char *dest, int size) = 0;
	virtual void rawRewind(bool start) {if (!start) ErrorHandler(EL_ERROR, "Can't loop sound since rawRewind() not implemented");}
};

extern struct SoundGroupSlots
{
	class Slot *slots;
	int active, max;
	bool is3d;
	float volume;

	SoundGroupSlots() : slots(NULL), active(0), volume(1) {}
	~SoundGroupSlots();
	void initialize(const struct SlotInitParams &sip);
} *soundGroups;//         NULL, ..     ,   

struct InitParams
{
	bool useHW, rightHandedCS;
	int channels, sampleRate, bitsPerSample, prefetchBytes;
	float bufferLength;//      ( )
	void *hwnd;//HWND    :(

	InitParams() : useHW(false), rightHandedCS(false), channels(2), sampleRate(44100), bitsPerSample(16), prefetchBytes(512*1024), bufferLength(.5f), hwnd(NULL) {}
};

struct SlotInitParams {int max; bool is3d;};
#define SG(name, max, is3d, ...) {max, is3d},
static const SlotInitParams defaultSlotsInitParams[SG_COUNT] = {S3_SOUND_GROUPS};
#undef  SG

bool Initialize(const InitParams &initPars = InitParams(), const SlotInitParams slotsIP[] = defaultSlotsInitParams, const int sgCount = SG_COUNT);
void Shutdown();
void SetMasterVolume(float f);
inline void SetSoundGroupVolume(SoundGroup sg, float volume) {if (soundGroups) soundGroups[sg].volume = volume;}
void SetListenerParameters(const float pos[3], const float front[3], const float top[3], const float vel[3] = NULL, float distFactor = 1, float rolloffFactor = 1, float dopplerFactor = 1);
}
