#include "stdafx.h"
#include "slot.h"


namespace s3
{
void Slot::createBuffer(const Format &f, bool is3d)
{
	format = f;

	DSBUFFERDESC dsbd = {sizeof(DSBUFFERDESC)};
	dsbd.dwFlags = DSBCAPS_CTRLVOLUME | DSBCAPS_GETCURRENTPOSITION2
		| (is3d ? DSBCAPS_CTRL3D : 0)
		| (!initParams.useHW ? DSBCAPS_LOCSOFTWARE : 0)
		| (!initParams.hwnd  ? DSBCAPS_GLOBALFOCUS : 0);
	WAVEFORMATEX wf = BuildWaveFormat(f.channels, f.sampleRate, f.bitsPerSample);
	dsbd.dwBufferBytes = int(initParams.bufferLength * wf.nAvgBytesPerSec);
	dsbd.dwBufferBytes += wf.nBlockAlign - dsbd.dwBufferBytes % wf.nBlockAlign;//   dwBufferBytes  wf.nBlockAlign    0
	dsbd.lpwfxFormat = &wf;
	bufferSize = dsbd.dwBufferBytes;

	LPDIRECTSOUNDBUFFER buf;
	if (FAILED(pDS->CreateSoundBuffer(&dsbd, &buf, NULL))) {ErrorHandler(EL_ERROR, "Can't create sound buffer"); return;}
	buf->QueryInterface(IID_IDirectSoundBuffer8, (LPVOID*)&buffer);//      
	if (is3d) buf->QueryInterface(IID_IDirectSound3DBuffer8, (LPVOID*)&buffer3D);
	buf->Release();
}

void Slot::startPlay(float time)
{
	buffer->GetCurrentPosition(&prevPlayPos, NULL);
	silenceSize = -1;
	paused = false;
	if (time) fade = 0, fadeSpeed = 1/time; else fade = 1, fadeSpeed = 0;
	update(0, true);
	buffer->Play(0, 0, DSBPLAY_LOOPING);
}

void Slot::read(void *buffer, int size, bool afterRewind)
{
	if (silenceSize == -1)
	{
		int r = std::max(decoder->read(buffer, size), 0);
		if (r != size)
		{
			if (looping && !(afterRewind && r == 0))//  -  ,    rewind  read  0 (   .) -    ,    
			{
				source->rewind(false);
				if (decoder->init(Source::readFunc, source))
					return read((char*)buffer + r, size - r, true);
			}

			silenceSize = 0;
			read((char*)buffer + r, size - r);
		}
	}
	else
	{
		memset(buffer, format.bitsPerSample == 8 ? 0x80 : 0, size);// 
		silenceSize += size;
	}
}

void Slot::stop()
{
	if (source == NULL) {ErrorHandler(EL_ERROR, "???"); return;}

	source->slotIndex = -1;
	source = NULL;
	buffer->Stop();
//	buffer->SetCurrentPosition(0);
}

void Slot::update(float dt, bool full)
{
	if (source == NULL) return;

	// 
	fade += fadeSpeed * dt;
	if (fade < 0) {stop(); return;}//    <= 0,  play()  time > 0    :), . if (time) fade = 0, ...
	if (fade > 1) fade = 1, fadeSpeed = 0;
	buffer->SetVolume(LinearToDirectSoundVolume(fade * source->volume * soundGroups[source->soundGroup].volume));

	//  
	DWORD playPos;
	if (SUCCEEDED(buffer->GetCurrentPosition(&playPos, NULL)))
	{
		DWORD readSize = (playPos >= prevPlayPos && !full) ? playPos - prevPlayPos : bufferSize - prevPlayPos + playPos;//     playPos > prevPlayPos,         (playPos == prevPlayPos),     

		LPVOID ptr1, ptr2;
		DWORD size1, size2;
		HRESULT hr = buffer->Lock(prevPlayPos, readSize, &ptr1, &size1, &ptr2, &size2, 0);

		if (hr == DSERR_BUFFERLOST)//   DX SDK: Using Streaming Buffers
		{
			buffer->Restore();
			hr = buffer->Lock(prevPlayPos, readSize, &ptr1, &size1, &ptr2, &size2, 0);
		}

		if (SUCCEEDED(hr))
		{
			read(ptr1, size1);
			if (ptr2) read(ptr2, size2);
			buffer->Unlock(ptr1, size1, ptr2, size2);

			prevPlayPos = playPos;

			if (silenceSize >= bufferSize) stop();//     ,  ,  play cursor        
		}
	}

	// 3D-
	if (buffer3D)
	{
		DS3DBUFFER pars;
 		pars.dwSize = sizeof(DS3DBUFFER);
 		ConvertVectorToDirectSoundCS(pars.vPosition, source->position);
 		ConvertVectorToDirectSoundCS(pars.vVelocity, source->velocity);
		pars.vConeOrientation.x = 0;
		pars.vConeOrientation.y = 0;
 		pars.vConeOrientation.z = 1;
		pars.dwInsideConeAngle = 360;
		pars.dwOutsideConeAngle = 360;
		pars.lConeOutsideVolume = 0;
		pars.flMinDistance = source->minDist;
		pars.flMaxDistance = source->maxDist;
		pars.dwMode = DS3DMODE_NORMAL;
		buffer3D->SetAllParameters(&pars, full/*       -    */ ? DS3D_IMMEDIATE : DS3D_DEFERRED);
	}
}
}
