#pragma once

#include "s3.h"
#include "decoder.h"


namespace s3
{
class Slot
{
public:
	Source *source;
	LPDIRECTSOUNDBUFFER8 buffer;
	LPDIRECTSOUND3DBUFFER8 buffer3D;
	DWORD prevPlayPos;
	Format format;
	int bufferSize, silenceSize;
	Decoder *decoder;// ,     std::swap

	bool paused, looping;
	float fade, fadeSpeed;

	Slot() : source(NULL), buffer(NULL), buffer3D(NULL), decoder(new Decoder) {}
	~Slot()
	{
		if (source) stop();
		delete decoder;
		releaseBuffer();
	}

	void releaseBuffer()
	{
		if (buffer  ) buffer  ->Release(), buffer   = NULL;
		if (buffer3D) buffer3D->Release(), buffer3D = NULL;
	}
	void createBuffer(const Format &f, bool is3d);
	void startPlay(float time);
	void stop();
	void stop(float time) {if (time == 0) stop(); else fadeSpeed = -1/time;}

	void read(void *buffer, int size, bool afterRewind = false);
	void update(float dt, bool full = false);
};

inline WAVEFORMATEX BuildWaveFormat(int channels, int sampleRate, int bitsPerSample)
{
	WAVEFORMATEX wf = {0};
	wf.wFormatTag = WAVE_FORMAT_PCM;
	wf.nChannels = (WORD)channels;
	wf.nSamplesPerSec = sampleRate;
	wf.wBitsPerSample = (WORD)bitsPerSample;
	wf.nBlockAlign = wf.nChannels * wf.wBitsPerSample / 8;
	wf.nAvgBytesPerSec = wf.nSamplesPerSec * wf.nBlockAlign;
	return wf;
}

inline int LinearToDirectSoundVolume(float k)
{
	return k <= 0 ? DSBVOLUME_MIN : k >= 1 ? DSBVOLUME_MAX : std::max(DSBVOLUME_MIN, int((100*10) * log10f(k)));//100 = -DSBVOLUME_MIN/100 dB, 10 -   dB (DirectSound is using dB for sound _power_)
}

extern InitParams initParams;
extern LPDIRECTSOUND8 pDS;
extern Decoder *gdecoder;

inline void ConvertVectorToDirectSoundCS(D3DVECTOR &dest, const float v[3])
{
	dest.x = v[0];
	if (initParams.rightHandedCS) dest.y = v[2], dest.z = v[1];//  -   z
	else						  dest.y = v[1], dest.z = v[2];
}
}
